import{m as p,l as f,i as w,b as e,r as x,o as F,_ as G,j as d,c as C,p as y}from"./index.72e24882.js";import{u as R,f as j}from"./Input.ed778d26.js";function M(o){return p("MuiFormGroup",o)}f("MuiFormGroup",["root","row","error"]);const S=["className","row"],U=o=>{const{classes:r,row:t,error:s}=o;return y({root:["root",t&&"row",s&&"error"]},M,r)},_=w("div",{name:"MuiFormGroup",slot:"Root",overridesResolver:(o,r)=>{const{ownerState:t}=o;return[r.root,t.row&&r.row]}})(({ownerState:o})=>e({display:"flex",flexDirection:"column",flexWrap:"wrap"},o.row&&{flexDirection:"row"})),g=x.exports.forwardRef(function(r,t){const s=F({props:r,name:"MuiFormGroup"}),{className:a,row:c=!1}=s,l=G(s,S),i=R(),u=j({props:s,muiFormControl:i,states:["error"]}),n=e({},s,{row:c,error:u.error}),m=U(n);return d.exports.jsx(_,e({className:C(m.root,a),ownerState:n,ref:t},l))}),N=g;export{N as F};
