import{r as e,n as a,S as r}from"./index.72e24882.js";import{B as n}from"./browse.0916686d.js";import{S as o}from"./default-highlight.19f6d178.js";import"./toPropertyKey.012fc217.js";const p=t=>t.json!==void 0?e.exports.createElement(n,null,e.exports.createElement(a,{id:"metadata-title",variant:"h6",component:"h2"},"Detailed Machine-Readable Representation of Item"),e.exports.createElement(o,{language:"json"},JSON.stringify(t.json,null,2))):e.exports.createElement(r,{variant:"rectangular"});export{p as default};
