{% extends "page.html" %}
{% block main %}
<div class = "container">
  <section class="section">
    <h1 class="title">Success</h1>
		<p>Return to your Tiled application. Within {{ interval }} seconds it should be succesfully logged in.</p>
  </section>
</div>
{% endblock %}
