import{r as e}from"./index.72e24882.js";import{B as l}from"./browse.0916686d.js";import{F as n}from"./FormLabel.a3eda031.js";import{I as s,S as i,M as m,F as p}from"./TextField.8d6efd58.js";const h=r=>{const a=Array.from(Array(r.npartitions).keys()),o=t=>{r.setValue(t.target.value)};return e.exports.createElement(l,null,e.exports.createElement(n,{sx:{my:2}},e.exports.createElement(s,{id:"partition-select-helper-label"},"Partition"),e.exports.createElement(i,{labelId:"partition-select-label",id:"partition-select",value:r.value,label:"Partition",onChange:o},a.map(t=>e.exports.createElement(m,{key:`partition-${t}`,value:t},t))),e.exports.createElement(p,null,"A portion of the rows")))};export{h as C};
