import{r as e,A as a,G as s,S as r}from"./index.72e24882.js";const n=e.exports.lazy(()=>a(()=>import("./array-1d.0797372d.js"),["assets/array-1d.0797372d.js","assets/index.72e24882.js","assets/index.d49bee11.css","assets/browse.0916686d.js","assets/index.esm.31c0e330.js","assets/Input.ed778d26.js","assets/FormLabel.a3eda031.js","assets/FormGroup.25250c88.js","assets/FormControlLabel.93f22896.js"])),c=e.exports.lazy(()=>a(()=>import("./array-nd.16ddeaf4.js"),["assets/array-nd.16ddeaf4.js","assets/index.72e24882.js","assets/index.d49bee11.css","assets/browse.0916686d.js","assets/index.esm.31c0e330.js","assets/Input.ed778d26.js"])),u=t=>{if(t.structure.data_type.hasOwnProperty("fields"))return e.exports.createElement(s,{severity:"warning"},'This is a "record array" with a'," ",e.exports.createElement("a",{href:"https://numpy.org/doc/stable/user/basics.rec.html",target:"_blank",rel:"noreferrer"},"structured data type"),'. The web interface cannot view it. Use the "Download" tab to access the data.');switch(t.structure.shape.length<2){case!0:return e.exports.createElement(e.exports.Suspense,{fallback:e.exports.createElement(r,{variant:"rectangular"})},e.exports.createElement(n,{segments:t.segments,item:t.item,link:t.link,structure:t.structure}));case!1:return e.exports.createElement(e.exports.Suspense,{fallback:e.exports.createElement(r,{variant:"rectangular"})},e.exports.createElement(c,{segments:t.segments,item:t.item,link:t.link,structure:t.structure}))}};export{u as default};
