import{i as I,b as n,r,o as x,_ as p,j as u,c as f,p as L}from"./index.72e24882.js";import{L as d,h as g}from"./TextField.8d6efd58.js";const S=["className"],h=s=>{const{alignItems:t,classes:e}=s;return L({root:["root",t==="flex-start"&&"alignItemsFlexStart"]},g,e)},v=I("div",{name:"MuiListItemIcon",slot:"Root",overridesResolver:(s,t)=>{const{ownerState:e}=s;return[t.root,e.alignItems==="flex-start"&&t.alignItemsFlexStart]}})(({theme:s,ownerState:t})=>n({minWidth:56,color:(s.vars||s).palette.action.active,flexShrink:0,display:"inline-flex"},t.alignItems==="flex-start"&&{marginTop:8})),C=r.exports.forwardRef(function(t,e){const o=x({props:t,name:"MuiListItemIcon"}),{className:i}=o,c=p(o,S),l=r.exports.useContext(d),a=n({},o,{alignItems:l.alignItems}),m=h(a);return u.exports.jsx(v,n({className:f(m.root,i),ownerState:a,ref:e},c))}),w=C;export{w as L};
