{% extends "page.html" %}
{% block main %}
<div class = "container">
  <section class="section">
    <h1 class="title">Authorize Tiled Session</h1>
		{% if message %}
		  <article class="message is-danger">
				<div class="message-header">
					<p>Error</p>
					<button class="delete" aria-label="delete"></button>
				</div>
				<div class="message-body">
					{{ message }}
				</div>
			</article>
		{% endif %}
    <form action="{{ action }}" method="post">
      <label for="user_code">Enter code</label>
      <input type="text" id="user_code" name="user_code" />
      <input type="hidden" id="code" name="code" value="{{ code }}" />
      <input type="submit" value="Enter" />
    </form>
  </section>
</div>
{% endblock %}
