<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{% block title %}Tiled{% endblock %}</title>
    {% block stylesheet %}
    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bulma@0.9.3/css/bulma.min.css">
    {% endblock %}
    {% block favicon %}
    {% endblock %}
    {% block scripts %}
    {% endblock %}
  </head>
  <body>
    {% block main %}
    {% endblock %}
  </body>
</html>
